
/* gnocl.q: Q-Gnocl interface and GQBuilder support library */

/* $Id: gnocl.q,v 1.5 2006/04/25 18:43:59 agraef Exp $ */

/* Interface. ***************************************************************/

/* Helper functions to quote Tk strings and escape Gnocl percent strings.
   These are useful in many Q-Gnocl applications so we provide them here. */

public tkstr S, percstr S;

/* app_path: This function resolves relative pathnames (absolute pathnames are
   always returned as is). We use this function internally to resolve relative
   image pathnames. It is provided here as a public function so that you can
   override it with a custom version when necessary, and use it to locate
   other data files of your application, too. The default version implemented
   in this module interprets relative pathnames w.r.t. the path of the main
   script, as given by which (), following symbolic links if necessary. This
   will probably be appropriate for most purposes. */

public app_path S;

/* gnocl_init: Perform the necessary initializations for Gnocl. This
   suppresses the Tk main window and loads the Gnocl extension packages. */

public gnocl_init;

/* The following "Widget" data type is used to represent GQBuilder widget tree
   descriptions. The main constructor "widget" has the arguments described
   below. (The "nowidget" constant is just a placeholder for missing paned
   children in GQBuilder and is replaced automatically with appropriate
   placeholder widgets when the widget tree is built with the gnocl_build
   function.)

   - NAME: The symbolic name of the widget. These are assumed to uniquely
     identify each widget in the tree (a condition which is actually enforced
     by GQBuilder), and can be used for the purpose of getting at the real
     Gnocl widget names once the widget tree has been processed with
     gnocl_build. (To these ends, gnocl_build returns a list of pairs of
     symbolic widget names and the corresponding Gnocl widget ids, see below.)

   - CLASS: The GQBuilder class name of the widget. This generally corresponds
     to the actual Gnocl command used to create the widget, except for the
     class buttonBox (which denotes a Gnocl box widget with the -buttonType
     option) and the toolBarItem classes (for which there actually are no
     widget creation commands in Gnocl; instead the widgets are created when
     they are added to the hosting toolbar widget).

   - DFLTS: The "widget defaults", i.e., the option values (list of option-
     value string pairs) with which the widget is created. Option names are
     specified without the leading dash. The widget defaults can normally be
     changed lateron by using the "configure" method on the widget (see
     gnocl_configure below).

   - CDFLTS: The "constrained defaults" of the widget. This is also a list of
     option-value pairs. However, these options are _not_ used at creation
     time, but when the widget is added to its parent. The available options
     thus depend on the parent and can _only_ be set when the child is added.
     Typical examples are alignment and padding options of children of box
     widgets.

   - CHILDREN: List of descriptions of the widget's children, if any.

   The option values used in the DFLTS and CDFLTS parameters usually
   correspond to the ranges of values for these options in Gnocl. One notable
   exception is the "mnemonicWidget" option of the "label" class which lets
   you specify the _symbolic_ name of the "buddy" widget, which will be
   resolved to the actual Gnocl widget id when the widget tree is processed.
   Moreover, gnocl_build also automatically resolves relative filenames for
   image and icon files, and translates GQBuilder synonyms for certain option
   values to the actual Gnocl/Tcl values (see the opt_val function in the
   implementation for details on this). Last but not least, when children are
   added to their parents, some CDFLTS option values are used as special
   arguments for the corresponding add method (see the description of
   gnocl_add below for details). */

public type Widget = const nowidget, widget NAME CLASS DFLTS CDFLTS CHILDREN;

/* is_widget: Check for Widget objects. */

public is_widget X;

is_widget _:Widget
		= true;
is_widget _	= false otherwise;

/* Convenience functions to access the components of a Widget object. */

public widget_name W, widget_class W, widget_defaults W,
  widget_cdefaults W, widget_children W;

widget_name (widget NAME _ _ _ _)
		= NAME;
widget_class (widget _ CLASS _ _ _)
		= CLASS;
widget_defaults (widget _ _ DFLTS _ _)
		= DFLTS;
widget_cdefaults (widget _ _ _ CDFLTS _)
		= CDFLTS;
widget_children (widget _ _ _ _ Ws)
		= Ws;

/* gnocl_build: Process widget descriptions and build the corresponding widget
   collection. The argument is either an object of the Widget type or a list
   of such objects. The function normally returns a list of name-id pairs of
   symbolic widget names and the corresponding Gnocl widget ids, which lists
   the widgets in the same order as in the widget tree, traversed in preorder.

   In particular, the topmost widget will always be the first in the result
   list, so you can easily retrieve it using the standard library function hd
   if the other widget ids are not required by the application. Otherwise it
   is common practice to convert the result list to a dictionary which allows
   easy access to the different widgets using the symbolic widget names as
   keys. (This is also the method employed in the GQBuilder default code
   template.)

   If the result of gnocl_build is not a list, then it is either an error
   message of the form tk_error MSG, indicating that something went wrong
   during widget instantiation, or some other result (e.g., in the case of
   modal toplevels it will be the return value of the dialog). */

public gnocl_build Ws;

/* Convenience functions to quickly create a single toplevel, common dialogs,
   and other widgets. */

public gnocl_dialog DFLTS, gnocl_window DFLTS;
public gnocl_file_chooser DFLTS;
public gnocl_color_selection DFLTS, gnocl_font_selection DFLTS;
public gnocl_widget CLASS DFLTS;

gnocl_dialog DFLTS:List
		= gnocl_widget "dialog" DFLTS;
gnocl_window DFLTS:List
		= gnocl_widget "window" DFLTS;
gnocl_file_chooser DFLTS:List
		= gnocl_widget "fileChooser" DFLTS;
gnocl_color_selection DFLTS:List
		= gnocl_widget "colorSelection" DFLTS;
gnocl_font_selection DFLTS:List
		= gnocl_widget "fontSelection" DFLTS;
private check_widget X;
gnocl_widget CLASS:String DFLTS:List
		= check_widget $ gnocl_build $ widget CLASS CLASS DFLTS [];
check_widget [(_,ID)]
		= ID;
check_widget X	= X otherwise;

/* gnocl_add: Add a list of widgets (or a single widget) as children to an
   existing parent widget. This provides a generic interface to add new
   children to all container widgets supported by GQBuilder, which hides away,
   as much as possible, the idiosyncrasies of the actual "add" methods
   involved for the different kinds of containers.

   To make this work, in addition to the Gnocl id of the parent you must also
   specify its GQBuilder class name so that the gnocl_add function can figure
   out the particular method to use for adding the children. The following
   parent classes are given special treatment and may allow or need special
   options in the CDFLTS argument of the child to supply the required
   additional parameters:

   - "window", "expander", "eventBox", "scrolledWindow": At most one child (of
     an arbitrary class provided that it is not a toplevel) can be added, which
     is configured using the "child" option of the parent.

   - "paned": At most two children are allowed, which are configured using the
     "children" option of the parent. Therefore all children _must_ be given
     in a single call to gnocl_add.

   - "menuBar", "menu": Children must be of one of the "menuItem" classes.

   - "menuItem": You can add a single "menu" child which provides a submenu
     under the menu item. This is achieved by configuring the "submenu" option
     of the parent.

   - "toolBar": Children must be of one of the "toolBarItem" classes or can be
     of any (non-toplevel) class. In the former case gnocl_add uses the
     appropriate variation of the toolbar's "add item" method, and you can
     supply the "icon" and "text" options in the child's CDFLTS to set the
     child's icon and text label as necessary. In the latter case, the
     toolbar's "add widget" method is used.

   - "notebook": In this case gnocl_add uses the notebook's addPage method to
     add the new children as additional notebook tabs. You can supply the
     "label" and "menu" options in the children's CDFLTS, which specify the
     labels shown on the tab respectively in the pages menu of the notebook
     widget.

   - "table": Along with additional options accepted by the table widget's add
     command, the children's CDFLTS can also supply the "row" and "column"
     options to specify the table cell of the child. (The default value for
     both "row" and "column" is "0".)

   - "canvas": The children's CDFLTS supplies the options for the canvas
     "create widget" command.

   For all other parent classes gnocl_add will use the generic "add" method of
   the parent to which it supplies the CDFLTS option values of the each child
   in turn. Note that in any case the CDFLTS option values can _not_ be
   changed once the children have been created and added.

   Like gnocl_build, the function normally returns a list of name-id pairs for
   the entire collection of newly created children. */

public gnocl_add ID CLASS Ws;

/* gnocl: Generic method invokation. This function allows you to invoke any
   method on a Gnocl widget. You have to specify the Gnocl id ID of the target
   widget, the method name METHOD (a string), the required parameters ARGS
   (either a singleton string or a tuple of strings) and the additional
   options OPTIONS (a list of option-value string pairs; as usual, option
   names are specified without the leading dash.). */

public gnocl ID METHOD ARGS OPTIONS;

/* Convenience functions to invoke the "delete", "configure" and "cget"
   methods. Note that, at the time of this writing (Gnocl 0.9.90), Gnocl
   supports the "cget" command only for a few widgets; please check the Gnocl
   documentation for details.  */

public gnocl_delete ID, gnocl_configure ID OPTIONS, gnocl_cget ID NAME;

gnocl_delete ID:String
		= gnocl ID "delete" () [];
gnocl_configure ID:String OPTIONS:List
		= gnocl ID "configure" () OPTIONS;
gnocl_configure ID:String (NAME:String,VAL:String)
		= gnocl ID "configure" () [(NAME,VAL)];
gnocl_cget ID:String NAME:String
		= gnocl ID "cget" ("-"++NAME) [];

/* gnocl_main: A simple main function to run an application. Takes as argument
   a list of either widget name-id pairs (as returned by gnocl_build) or just
   plain Gnocl widget ids, or a single such item, and installs an onDestroy
   handler on all given toplevel windows (unless they already have one) which
   exits the main program. The function then enters the Gnocl main loop (which
   never returns), unless there were no visible toplevel windows, in which
   case gnocl_main just fails. */

public gnocl_main IDS;

/* gnocl_popup: A simple dialog shell, which lets you turn any toplevel window
   into a dialog. It takes as input an arbitrary window TOPLEVEL (a Gnocl id),
   a parent window PARENT (this one may also be ()), a list of button
   definitions BUTTONS and a list of key bindings KEYS. If PARENT is not ()
   and a valid toplevel, the geometry of TOPLEVEL is set up (unless it already
   has a display position other than (0,0)) so that the window is centered
   above PARENT. The gnocl_popup function then makes TOPLEVEL visible and, if
   TOPLEVEL is a modal window, processes events until the user either closes
   the window, pushes one of the given buttons, or presses a key which in turn
   causes one of the given buttons to be activated, in which case TOPLEVEL is
   made invisible and a value is returned to the caller.

   Each entry of the BUTTONS list must be of the form (ID,CB,VAL) where ID is
   the Gnocl id of the button (or any other widget which provides an onClicked
   callback), CB is a Q function to be invoked when the button is clicked, and
   VAL is the value to be supplied to CB when the button is clicked. (Currently,
   both CB and VAL must be serializable with str, as they are stored as a
   callback on the button widget.) Note that gnocl_popup does not actually
   create any buttons itself, so they must already be in the dialog toplevel;
   however, it will set the onClicked callbacks of the given buttons to
   perform the appropriate actions which close the dialog and return a value
   to the caller.

   Each entry of the KEYS list is a key binding of the form (KEY,ID) or
   (KEY,SHIFT,ID), where KEY is the symbolic name of a key, SHIFT (optionally)
   is an integer bitmask specifying a shift state, and ID is the Gnocl id of a
   button which is to be activated when the user presses the key in the given
   shift state (or in any shift state if none is specified). The keybindings
   will be assigned to the toplevel window; to these ends, the toplevel's
   onKeyPress callback is redefined accordingly. Typically KEYS is used to
   specify bindings for special keys such as Return, Escape, etc., which are
   equivalent to button clicks, but it may actually specify assignments for
   any widget which can execute the onClicked command, even widgets which are
   not in the BUTTONS list.

   A modal dialog returns the result of applying the callback CB of the
   activated button to the corresponding value VAL, or () if the window was
   closed from the window manager. In the case of a non-modal dialog
   gnocl_popup returns TOPLEVEL immediately after the dialog has been made
   visible. */

public gnocl_popup TOPLEVEL PARENT BUTTONS KEYS;

/* Implementation. **********************************************************/

import tk;

/* Quote Tcl/Tk strings. */

private tkch C;
tkch "{"	= "\\{";
tkch "}"	= "\\}";
tkch "["	= "\\[";
tkch "]"	= "\\]";
tkch "$"	= "\\$";
tkch "\\"	= "\\\\";
tkch "\""	= "\\\"";
tkch C		= C otherwise;

tkstr S:String	= sprintf "\"%s\"" (strcat (map tkch (chars S)));

/* Quote Gnocl percent strings ("%#" etc.) if necessary. */

percstr S:String
		= "%"++S if not null S and then (S!0="%");
		= S otherwise;

/* Split a percent string into prefix and main part. */

private splitperc S;
splitperc S:String
		= (PS,S)
		    if (#PS=2) and then (PS!0="%") and then
		      (pos (PS!1) "!%_#</">=0)
		    where (PS:String,S:String) =
		      (sub S 0 1,sub S 2 (#S-1));

/* Resolve relative pathnames and links. */

// normalize a path by removing "." and ".." directories
private pushdir Ps P, normalize Ps;
pushdir Ps "."	= Ps;
pushdir [] ".."	= [];
pushdir [P|Ps] ".."
		= Ps;
pushdir Ps P	= [P|Ps] otherwise;
normalize Ps	= reverse $ foldl pushdir [] Ps;

// dirname part of a filename
private dirname PATH;
dirname ""	= "";
dirname "/"	= "/";
dirname PATH	= join "/" (init $ split "/" PATH) ++ "/"
		    otherwise;

// absolute pathname relative to given dir
private prefix DIR, absname DIR PATH, abslink PATH LINK;
prefix DIR	= DIR if DIR="/";
		= DIR++"/" otherwise;
absname DIR PATH
		= join "/" $ cons "" $ normalize $ filter (neg null) $
		  split "/" $ prefix DIR++PATH
		    if not null PATH and then (PATH!0<>"/");
		= PATH otherwise;
abslink PATH LINK
		= absname (dirname PATH) LINK;

// follow a sequence of links to the original file, avoid looping
private xlink S T, realname S;
xlink S S	= S;
xlink S T	= xlink S T
		    where T:String = readlink T, T = abslink S T;
		= T otherwise;
realname S	= xlink S T
		    where T:String = readlink S, T = abslink S T;
		= S otherwise;

// the load path of the main script
private loadpath;
loadpath	= dirname (realname WHICH) if not null WHICH
		    where WHICH:String = which ();
		= "" otherwise;

def PREFIX = loadpath;

app_path S:String
		= join "/" $ cons "" $ normalize $ filter (neg null) $
		  split "/" $ PREFIX++S
		    if not null S and then (S!0<>"/");
		= S otherwise;

/* Check string values, map to default if necessary. */

private checkval X Y;
checkval _ X:String
		= X if not null X;
checkval X _	= X otherwise;

/* The real Gnocl classname. */

private classname CLASS;
classname "buttonBox"
		= "box";
classname CLASS	= CLASS otherwise;

/* Translate relative image pathnames, as well as GQBuilder synonyms for
   certain option values to the actual Gnocl/Tcl values. */

private opt_val OPT VAL;
opt_val "icon" IMG
		= "%/"++app_path IMG where ("%/",IMG) = splitperc IMG;
opt_val "image" IMG
		= "%/"++app_path IMG where ("%/",IMG) = splitperc IMG;

opt_val "fill" "none"
		= "0 0";
opt_val "fill" "both"
		= "1 1";
opt_val "fill" "x"
		= "1 0";
opt_val "fill" "y"
		= "0 1";
opt_val "expand" "none"
		= "1 1";
opt_val "expand" "both"
		= "1 1";
opt_val "expand" "x"
		= "1 0";
opt_val "expand" "y"
		= "0 1";
// opt_val _ "false"
// 		= "0";
// opt_val _ "true"
// 		= "1";

opt_val _ VAL	= VAL otherwise;

/* Build an option string from an attribute-value pair. */

private optstr REF NAME X;
optstr REF:Ref NAME:String ("mnemonicWidget",VAL:String)
		= put REF $ [(NAME,("mnemonicWidget",VAL))|get REF] || "";
optstr _ _ (NAME:String,VAL:String)
		= sprintf "-%s %s" (NAME,tkstr $ opt_val NAME VAL);
optstr _ _ (OPTNAME:String,tk_error MSG)
		= throw $ tk_error MSG;
optstr _ NAME (OPTNAME:String,VAL)
		= throw $ tk_error $ sprintf "%s: %s"
		  (NAME,sprintf "bad value '%s' for option %s"
		   (str VAL,str OPTNAME))
		    if isstr NAME;
		= throw $ tk_error $ sprintf "bad value '%s' for option %s"
		  (str VAL,str OPTNAME)
		    otherwise;
optstr _ NAME X	= throw $ tk_error $ sprintf "%s: %s"
		  (NAME,sprintf "bad option %s" $ str X)
		    if isstr NAME;
		= throw $ tk_error $ sprintf "bad option %s" $ str X
		    otherwise;

/* Error checker. */

private check P X Y;
check P _ X	= X if P X;
check _ NAME:String (tk_error MSG)
		= throw $ tk_error $ sprintf "%s: %s" (NAME,MSG)
		    if not null NAME;
check _ _ X	= throw X otherwise;

/* Widget instantiation. */

private create REF NAME CLASS DFLTS;
create REF NAME CLASS DFLTS
		= check isstr NAME $ tk $ sprintf "gnocl::%s %s"
		  (classname CLASS,join " " $ map (optstr REF NAME) DFLTS);

/* Add a child to its parent. Special cases here: window/expander/eventBox/
   scrolledWindow (single child configured with -child), menuItem (single menu
   child configured with submenu), notebook (page and optional page menu label
   argument, inserted with addPage), table (additional column and row
   arguments). The default case just uses the add method with CDFLTS as
   options. */

private isstr_or_null X;
isstr_or_null X	= isstr X or else eq () X;

private add REF ID CLASS W;
add REF ID CLASS (widget (_,CID) CCLASS _ _ _)
		= tk $ sprintf "%s configure -child %s" (ID,CID)
		    if any (=CLASS)
		      ["window","expander","eventBox","scrolledWindow"];
add REF ID "menuItem" (widget (_,CID) CCLASS _ _ _)
		= tk $ sprintf "%s configure -submenu %s" (ID,CID);
add REF ID "notebook" (widget (_,CID) CCLASS _ CDFLTS _)
		= tk $ sprintf "%s addPage %s %s %s" (ID,CID,LABEL,MENU)
		    where CDFLTS = dict CDFLTS,
		      LABEL = tkstr $ checkval "" $ CDFLTS!"label",
		      MENU = checkval "" $ CDFLTS!"menu",
		      MENU = ifelse (null MENU) "" (tkstr MENU);
add REF ID "table" (widget (NAME,CID) CCLASS _ CDFLTS _)
		= check isstr_or_null NAME $
		  tk $ sprintf "%s add %s %s %s %s" (ID,CID,COLUMN,ROW,OPTIONS)
		    where CDFLTS = dict CDFLTS,
		      ROW = tkstr $ checkval "0" $ CDFLTS!"row",
		      COLUMN = tkstr $ checkval "0" $ CDFLTS!"column",
		      CDFLTS = list $ foldl delete CDFLTS ["row","column"],
		      OPTIONS = join " " $ map (optstr REF NAME) CDFLTS;
add REF ID "canvas" (widget (NAME,CID) CCLASS _ CDFLTS _)
		= check isstr_or_null NAME $
		  tk $ sprintf "%s create widget -widget %s %s"
		  (ID,CID,join " " $ map (optstr REF NAME) CDFLTS);
add REF ID CLASS (widget (NAME,CID) CCLASS _ CDFLTS _)
		= check isstr_or_null NAME $
		  tk $ sprintf "%s add %s %s"
		  (ID,CID,join " " $ map (optstr REF NAME) CDFLTS)
		    otherwise;

/* Special case: add paned children (these must all be configured at once). */

private widget_id W, paned_child N W, paned_children Ws,
  add_paned_children ID CWs;

widget_id (widget (_,ID) _ _ _ _)
		= ID;

paned_child N nowidget
		= widget "" "label" [("text",sprintf "Pane#%d" N)] [] [];
paned_child _ W	= W otherwise;

paned_children [W]
		= paned_children [W,nowidget];
paned_children Ws
		= zipwith paned_child [1..#Ws] Ws otherwise;

add_paned_children ID CWs
		= () if null CWs;
		= tk $ sprintf "%s configure -children {%s}"
		  (ID,join " " $ map widget_id CWs);

/* Special case: add toolbar items. */

private toolbar_item CLASS, add_toolbar_item REF ID W;

toolbar_item "toolBarItem"
		= "item";
toolbar_item "toolBarCheckItem"
		= "checkItem";
toolbar_item "toolBarRadioItem"
		= "radioItem";
toolbar_item "toolBarSpace"
		= "space";

add_toolbar_item REF ID (widget NAME CLASS DFLTS CDFLTS CWs)
		= widget (NAME,CID) CLASS DFLTS CDFLTS CWs
		    where CID = check isstr_or_null NAME $
		      tk $ sprintf "%s add %s %s"
		      (ID,toolbar_item CLASS,
		       join " " $ map (optstr REF NAME) $ CDFLTS++DFLTS)
		    if any (=CLASS) ["toolBarItem","toolBarCheckItem",
				     "toolBarRadioItem","toolBarSpace"];

private build REF W;

add_toolbar_item REF ID W
		= W where W = build REF W,
		      widget (_,CID) _ _ CDFLTS _ = W,
		      CID1 = tk $ sprintf "%s add widget %s" (ID,CID)
		      // assert: CID1=CID
		    otherwise;

/* Recursively build the widget tree from a corresponding Widget object.
   Special cases here: toolBar (children are created and added in a single
   step), paned (filter out nowidget children), notebook (page option must be
   set _after_ the children have been created). */

build REF (widget NAME CLASS DFLTS CDFLTS CWs)
		= widget (NAME,ID) CLASS DFLTS CDFLTS CWs
		    where ID = create REF NAME CLASS DFLTS,
		      CWs = map (add_toolbar_item REF ID) CWs
		    if CLASS="toolBar";
build REF (widget NAME CLASS DFLTS CDFLTS CWs)
		= do (add REF ID CLASS) CWs ||
		  when (isstr PAGE) $ gnocl_configure ID [("page",PAGE)] ||
		  widget (NAME,ID) CLASS DFLTS CDFLTS CWs
		    where MYDFLTS = dict DFLTS, PAGE = MYDFLTS!"page",
		      MYDFLTS = list $ delete MYDFLTS "page",
		      ID = create REF NAME CLASS MYDFLTS,
		      CWs = map (build REF) CWs
		    if CLASS="notebook";
build REF (widget NAME CLASS DFLTS CDFLTS CWs)
		= check isstr_or_null NAME $ add_paned_children ID CWs ||
		  widget (NAME,ID) CLASS DFLTS CDFLTS CWs
		    where ID = create REF NAME CLASS DFLTS,
		      CWs = map (build REF) $ paned_children CWs
		    if CLASS="paned";
build REF (widget NAME CLASS DFLTS CDFLTS CWs)
		= do (add REF ID CLASS) CWs ||
		  widget (NAME,ID) CLASS DFLTS CDFLTS CWs
		    where ID = create REF NAME CLASS DFLTS,
		      CWs = map (build REF) CWs
		    otherwise;
build _ W	= throw W otherwise;

/* Create widget lists and dictionaries from the name-id pairs. */

private widget_accum Xs Ws, widget_list Ws, widget_dict Ws;

widget_accum Xs Ws:List
		= foldl widget_accum Xs Ws;
widget_accum Xs (widget (NAME,ID) CLASS DFLTS CDFLTS CWs)
		= widget_accum [(NAME,ID)|Xs] CWs;
widget_accum Xs X
		= throw X otherwise;

widget_list Ws	= reverse $ foldl widget_accum [] Ws;

widget_dict Ws	= dict $ filter (isstr.fst) Ws;

/* Resolve mnemonicWidget references. */

private resolve WDICT NAME OPTION;
resolve WDICT NAME1 (OPT,NAME2)
		= gnocl_configure (WDICT!NAME1) [(OPT,NAME2)]
		    where NAME2:String = WDICT!NAME2;
		= gnocl_configure (WDICT!NAME1) [(OPT,NAME2)]
		    otherwise;

/* Wrapper routines which do the grunt work for gnocl_build and gnocl_add. */

private test_build Ws, test_add1 REF ID CLASS W, test_add ID CLASS Ws;

test_build Ws	= do (uncurry (resolve $ widget_dict Ws)) $ get REF || Ws
		    where REF = ref [],
		      Ws = widget_list $ map (build REF) Ws;

test_add ID CLASS Ws
		= do (uncurry (resolve $ widget_dict Ws)) $ get REF || Ws
		    where REF = ref [],
		      Ws = map (build REF) $ paned_children Ws,
		      _ = check isstr_or_null () $ add_paned_children ID Ws,
		      Ws = widget_list Ws
		    if CLASS="paned";
		= do (uncurry (resolve $ widget_dict Ws)) $ get REF || Ws
		    where REF = ref [],
		      Ws = widget_list $
		      map (test_add1 REF ID CLASS) Ws;

test_add1 REF ID CLASS W
		= add_toolbar_item REF ID W
		    where NAME = widget_name W
		    if CLASS="toolBar";
		= check isstr_or_null NAME $ add REF ID CLASS CW || CW
		    where NAME = widget_name W,
		      CW = build REF W
		    otherwise;

/* Interface functions. */

/* gnocl_init */

gnocl_init	= tk "wm withdraw ." || do (tk.sprintf "package require %s")
		  ["Gnocl","GnoclCanvas","GnoclGconf"];

/* gnocl_build and gnocl_add: These are now simple wrappers which just catch
   and return exception values. */

gnocl_build W:Widget
		= gnocl_build [W];
gnocl_build Ws:List
		= catch id $ test_build Ws;

gnocl_add ID:String CLASS:String W:Widget
		= gnocl_add ID CLASS [W];
gnocl_add ID:String CLASS:String Ws:List
		= catch id $ test_add ID CLASS Ws;

/* gnocl */

gnocl ID:String METHOD:String ARG:String OPTIONS:List
		= gnocl ID METHOD (ARG,) OPTIONS;
gnocl ID:String METHOD:String ARGS:Tuple OPTIONS:List
		= catch fail $ tk $ sprintf "%s %s %s %s"
		  (ID,METHOD,ARGS,join " " $ map (optstr () ID) OPTIONS)
		    where ARGS:String = join " " $ map tkstr $ list ARGS;

/* gnocl_main */

private is_visible X, can_have_handler X, must_have_handler X, myid ID;

is_visible (tk_error _)
		= false;
is_visible X	= eq X () or else eq X "1" otherwise;

can_have_handler (tk_error _)
		= false;
can_have_handler X
		= isstr X or else eq () X otherwise;

must_have_handler (tk_error _)
		= false;
must_have_handler X:String
		= null X;
must_have_handler _
		= true otherwise;

myid (NAME,ID)	= ID;
myid ID		= ID otherwise;

gnocl_main ID:String
		= gnocl_main [ID];
gnocl_main (_,ID:String)
		= gnocl_main [ID];
gnocl_main IDS:List
		= do (tk.sprintf "%s configure -onDestroy {exit 0}") IDS ||
		  tk "gnocl::mainLoop"
  where IDS = filter (must_have_handler.tk.sprintf "%s cget -onDestroy") IDS
  if not null IDS
  where IDS = map myid IDS,
    IDS = filter (can_have_handler.tk.sprintf "%s cget -onDestroy") IDS,
    IDS = filter (is_visible.tk.sprintf "%s cget -visible") IDS;

/* gnocl_popup */

private dg_popdown DG, dg_close_cb DG;

dg_popdown DG	= gnocl_configure DG ("visible","0");

dg_close_cb DG	= dg_popdown DG || "0";

public find_key KEYS KEY SHIFT, dg_keypress_cb KEYS KEY SHIFT;

find_key [(KEY,ID)|KEYS] KEY SHIFT
		= ID;
find_key [(KEY,S,ID)|KEYS] KEY SHIFT
		= ID if (S and SHIFT)<>0;
find_key [_|KEYS] KEY SHIFT
		= find_key KEYS KEY SHIFT
		    otherwise;

dg_keypress_cb KEYS KEY SHIFT
		= tk $ sprintf "%s onClicked" ID
		    where ID:String = find_key KEYS KEY SHIFT;

private dg_button_cb DG CB VAL, dg_button MODAL DG BUTTONS;

dg_button_cb DG CB VAL
		= dg_popdown DG || CB VAL;

dg_button MODAL DG (ID,CB,VAL)
		= gnocl_configure ID
		  [("onClicked",CMD++" "++
		    (tkstr $ sprintf "gnocl::dg_button_cb %s (%s) (%s)"
		     (str DG,CB,VAL)))]
		    where CMD = ifelse MODAL "q" "qval";

private dg_main_loop DG, dg_flush, dg_main MODAL DG;

dg_main_loop DG	= tk_read if tk "gnocl::update" || tk_check;
		= sleep 0.1 || dg_main_loop DG
		    where "1" = gnocl_cget DG "visible";
		= () otherwise;

dg_flush	= tk_reads || dg_flush if tk_check;

dg_main MODAL DG
		= dg_flush || dg_main_loop DG if MODAL;
		= DG otherwise;

private geometry ID, centerpos ID CDIM;

geometry ID	= (W,H,X,Y)
		    where [W:Int,H:Int,X:Int,Y:Int] =
		      map (val.tk.sprintf (sprintf "%s cget -%%s" ID))
		      ["width","height","x","y"];

centerpos ID (CW,CH)
		= (max 0 X,max 0 Y)
		    where (W,H,X,Y) = geometry ID,
		      X = X+(W-CW) div 2, Y = Y+(H-CH) div 2;

private dg_popup TOPLEVEL BUTTONS KEYS;

dg_popup TOPLEVEL:String BUTTONS:List KEYS:List
		= // reconfigure toplevel and buttons
		  gnocl_configure TOPLEVEL
		  [("visible","1"),
		   ("onDelete",
		    sprintf "qval {gnocl::dg_close_cb %s}" (str TOPLEVEL)),
		   ("onKeyPress",
		    sprintf "qval {gnocl::dg_keypress_cb %s %s %%s}"
		    (str KEYS,str "%K"))] ||
		  do (dg_button MODAL TOPLEVEL) BUTTONS ||
		  // invoke the main loop
		  dg_main MODAL TOPLEVEL
		    where // check whether we're running as a modal dialog
		      MODAL = eq "1" $ gnocl_cget TOPLEVEL "modal";

gnocl_popup TOPLEVEL:String PARENT:String BUTTONS:List KEYS:List
		= // centered position on the main window by default,
		  // otherwise we keep the position as the user left it
		  // the last time the dialog was visible
		  unless (all isint [X,Y] and then any (>0) [X,Y]) $
		  gnocl_configure TOPLEVEL [("x",str X0),("y",str Y0)] ||
		  dg_popup TOPLEVEL BUTTONS KEYS
		    where [W,H,X,Y] = map (val.gnocl_cget TOPLEVEL)
		      ["width","height","x","y"],
		      (X0,Y0) = centerpos PARENT (W,H);

gnocl_popup TOPLEVEL:String () BUTTONS:List KEYS:List
		= dg_popup TOPLEVEL BUTTONS KEYS otherwise;
